"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var logger_1 = require("../../../../utils/logger");
var intents_android_1 = require("./intents.android");
var events_1 = require("../../../../events");
var time_1 = require("../../../../utils/time");
var TIMEOUT = 180000;
var TIMEOUT_EVENT_OFFSET = 5000;
var IDLE_TIME = 1000;
var TaskChainRunnerService = (function () {
    function TaskChainRunnerService() {
    }
    TaskChainRunnerService.prototype.onCreate = function (nativeService) {
        this.nativeService = nativeService;
        this.wakeLock = taskChainRunnerWakeLock(nativeService);
        this.wakeLock.setReferenceCounted(false);
        this.wakeLock.acquire();
        this.timeoutIds = new Set();
        this.taskChainCount = 0;
        this.killed = false;
        this.logger = logger_1.getLogger("TaskChainRunnerService");
        this.logger.debug("onCreate called");
        this.initializeExecutionWindow();
    };
    TaskChainRunnerService.prototype.onStartCommand = function (intent, flags, startId) {
        var _this = this;
        this.logger.info("Service called {flags=" + flags + ", startId=" + startId + "}");
        if (intent) {
            this.taskChainStarted();
            this.processRequest(intent)
                .then(function () {
                _this.logger.debug("Task chain finished running");
                _this.taskChainDone();
            })
                .catch(function (err) {
                _this.logger.error("Error while executing the task chain: " + err);
                _this.taskChainDone();
            });
        }
        return android.app.Service.START_REDELIVER_INTENT;
    };
    TaskChainRunnerService.prototype.onDestroy = function () {
        this.logger.debug("onDestroy called");
    };
    TaskChainRunnerService.prototype.initializeExecutionWindow = function () {
        this.executionStart = time_1.now();
        this.wakeLock.acquire(TIMEOUT);
    };
    TaskChainRunnerService.prototype.processRequest = function (intent) {
        return __awaiter(this, void 0, void 0, function () {
            var params, _a, launchEvent, timeoutId;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        params = this.extractRequestParams(intent);
                        _a = __read(this.prepareTaskChainExecution(params), 2), launchEvent = _a[0], timeoutId = _a[1];
                        return [4, this.startTaskChainExecution(launchEvent, timeoutId)];
                    case 1:
                        _b.sent();
                        return [4, this.stayIdleIfPossible(IDLE_TIME)];
                    case 2:
                        _b.sent();
                        return [2];
                }
            });
        });
    };
    TaskChainRunnerService.prototype.taskChainStarted = function () {
        this.taskChainCount++;
    };
    TaskChainRunnerService.prototype.taskChainDone = function () {
        this.taskChainCount--;
        if (this.taskChainCount === 0) {
            this.gracefullyStop();
        }
    };
    TaskChainRunnerService.prototype.extractRequestParams = function (intent) {
        var params = intents_android_1.unpackTaskChainRunnerServiceIntent(intent);
        this.logger.info("Request params {launchEvent=" + params.launchEvent + ", eventData=" + JSON.stringify(params.eventData) + (params.eventId ? ", eventId=" + params.eventId : "") + "}");
        return params;
    };
    TaskChainRunnerService.prototype.prepareTaskChainExecution = function (params) {
        var _this = this;
        var timeout = this.calculateTimeout();
        var launchEvent = events_1.createEvent(params.launchEvent, {
            data: params.eventData,
            expirationTimestamp: this.getExpirationTimestamp(timeout),
        });
        if (params.eventId) {
            launchEvent.id = params.eventId;
        }
        var id = launchEvent.id;
        var timeoutEvent = events_1.createEvent(events_1.TaskDispatcherEvent.TaskExecutionTimedOut, {
            id: id,
        });
        this.logger.info("Execution will timeout in " + timeout + ", for tasks running with execution id: " + id);
        var timeoutId = setTimeout(function () {
            events_1.emit(timeoutEvent);
            _this.logger.warn("Execution timed out for tasks running with execution id: " + id);
        }, timeout);
        this.timeoutIds.add(timeoutId);
        return [launchEvent, timeoutId];
    };
    TaskChainRunnerService.prototype.startTaskChainExecution = function (launchEvent, timeoutId) {
        var _this = this;
        return new Promise(function (resolve) {
            var listenerId = events_1.on(events_1.TaskDispatcherEvent.TaskChainFinished, function (evt) {
                if (evt.id === launchEvent.id) {
                    events_1.off(events_1.TaskDispatcherEvent.TaskChainFinished, listenerId);
                    _this.logger.info("Task chain with id (" + launchEvent.id + ") finished its execution");
                    _this.timeoutIds.delete(timeoutId);
                    clearTimeout(timeoutId);
                    resolve();
                }
            });
            events_1.emit(launchEvent);
        });
    };
    TaskChainRunnerService.prototype.stayIdleIfPossible = function (idleTime) {
        var timeout = this.calculateTimeout();
        if (timeout < idleTime) {
            return Promise.resolve();
        }
        return new Promise(function (resolve) { return setTimeout(resolve, idleTime); });
    };
    TaskChainRunnerService.prototype.calculateTimeout = function () {
        var currentMillis = time_1.now();
        var diff = currentMillis - this.executionStart;
        return TIMEOUT - TIMEOUT_EVENT_OFFSET - diff;
    };
    TaskChainRunnerService.prototype.getExpirationTimestamp = function (timeout) {
        return time_1.now() + timeout;
    };
    TaskChainRunnerService.prototype.gracefullyStop = function () {
        var e_1, _a;
        this.logger.debug("Stopping service");
        this.killWithFire();
        try {
            for (var _b = __values(this.timeoutIds), _c = _b.next(); !_c.done; _c = _b.next()) {
                var timeoutId = _c.value;
                clearTimeout(timeoutId);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        if (this.wakeLock.isHeld()) {
            this.wakeLock.release();
            this.logger.debug("Lock released");
        }
    };
    TaskChainRunnerService.prototype.killWithFire = function () {
        var startId = 1;
        while (!this.killed) {
            var last = this.nativeService.stopSelfResult(startId);
            if (last) {
                this.killed = true;
            }
            else {
                startId++;
            }
        }
        this.logger.info("Done running (startId=" + startId + ")");
    };
    return TaskChainRunnerService;
}());
exports.TaskChainRunnerService = TaskChainRunnerService;
function taskChainRunnerWakeLock(context) {
    var wakeLockName = "TaskDispatcher::TaskChainRunnerWakeLock";
    var powerManager = context.getSystemService(android.content.Context.POWER_SERVICE);
    return powerManager.newWakeLock(android.os.PowerManager.PARTIAL_WAKE_LOCK, wakeLockName);
}
var _taskChainRunnerService;
function getTaskChainRunnerService() {
    if (!_taskChainRunnerService) {
        _taskChainRunnerService = new TaskChainRunnerService();
    }
    return _taskChainRunnerService;
}
exports.getTaskChainRunnerService = getTaskChainRunnerService;
//# sourceMappingURL=runner-service.android.js.map