"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var application_1 = require("tns-core-modules/application/application");
var intents_android_1 = require("../../intents.android");
var notification_manager_android_1 = require("../../notification-manager.android");
var planned_tasks_store_1 = require("../../../../../../persistence/planned-tasks-store");
var batch_task_runner_1 = require("../../../../../runners/batch-task-runner");
var events_1 = require("../../../../../../events");
var manager_1 = require("../../../../../manager");
var planned_task_1 = require("../../../../../planner/planned-task");
var logger_1 = require("../../../../../../utils/logger");
var time_1 = require("../../../../../../utils/time");
var foreground_checker_1 = require("../../../../../foreground-checker");
var MIN_TIMEOUT = 60000;
var TIMEOUT_EVENT_OFFSET = 5000;
var AlarmRunnerService = (function () {
    function AlarmRunnerService() {
    }
    AlarmRunnerService.prototype.onCreate = function (nativeService) {
        this.nativeService = nativeService;
        this.runsInForeground = false;
        this.timeOffset = 0;
        this.currentTime = time_1.now();
        this.started = false;
        this.inForeground = false;
        this.wakeLock = alarmRunnerWakeLock(nativeService);
        this.wakeLock.acquire();
        this.wakeLock.setReferenceCounted(false);
        this.taskStore = planned_tasks_store_1.plannedTasksDB;
        this.foregroundChecker = new foreground_checker_1.ForegroundChecker();
        this.logger = logger_1.getLogger("AlarmRunnerService");
        this.logger.debug("onCreate called");
    };
    AlarmRunnerService.prototype.onStartCommand = function (intent, flags, startId) {
        var _this = this;
        this.logger.info("Service called {flags=" + flags + ", startId=" + startId + "}");
        var startFlag = android.app.Service.START_REDELIVER_INTENT;
        if (this.alreadyRunning(startId)) {
            return startFlag;
        }
        this.extractInvocationArguments(intent);
        if (this.runsInForeground) {
            this.runInForeground();
        }
        this.runTasks()
            .then(function () {
            _this.logger.debug("Tasks finished running");
            _this.gracefullyStop();
        })
            .catch(function (err) {
            _this.logger.error("Error while running tasks " + err);
            _this.gracefullyStop();
        });
        return startFlag;
    };
    AlarmRunnerService.prototype.onDestroy = function () {
        this.logger.debug("onDestroy called");
        this.gracefullyStop();
        this.nativeService = null;
    };
    AlarmRunnerService.prototype.alreadyRunning = function (startId) {
        if (startId === 1) {
            this.started = true;
            this.logger.debug("Service started");
            return false;
        }
        this.logger.warn("Service already running! Dismissing call -> startId: " + startId + "!");
        this.nativeService.stopSelf(startId);
        return true;
    };
    AlarmRunnerService.prototype.extractInvocationArguments = function (intent) {
        var args = intents_android_1.unpackAlarmRunnerServiceIntent(intent);
        this.logger.info("InvocationParams {runsInForeground=" + args.runInForeground + ", offset=" + args.timeOffset + ", invocationTime=" + args.invocationTime + "}");
        this.runsInForeground = args.runInForeground;
        this.timeOffset = args.timeOffset;
        this.currentTime = args.invocationTime;
    };
    AlarmRunnerService.prototype.runInForeground = function () {
        if (this.inForeground) {
            return;
        }
        this.nativeService.startForeground(notification_manager_android_1.AndroidNotification.LocationUsage, notification_manager_android_1.createNotification(this.nativeService, notification_manager_android_1.AndroidNotification.LocationUsage));
        this.inForeground = true;
        this.logger.debug("Running in foreground");
    };
    AlarmRunnerService.prototype.runTasks = function () {
        return __awaiter(this, void 0, void 0, function () {
            var taskManager, tasksToRun, taskCount, timeout, executionStartedEvt, taskRunner;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        taskManager = this.createTaskManager();
                        return [4, taskManager.tasksToRun()];
                    case 1:
                        tasksToRun = _a.sent();
                        taskCount = tasksToRun.length;
                        if (!(taskCount > 0)) return [3, 4];
                        return [4, this.calculateTimeout(taskManager)];
                    case 2:
                        timeout = _a.sent();
                        executionStartedEvt = this.initializeExecutionWindow(timeout);
                        taskRunner = new batch_task_runner_1.BatchTaskRunner(this.taskStore);
                        this.logger.info("Running " + taskCount + " tasks");
                        return [4, taskRunner.run(tasksToRun, executionStartedEvt)];
                    case 3:
                        _a.sent();
                        return [3, 5];
                    case 4:
                        this.logger.warn("Service was called but no tasks were run!");
                        _a.label = 5;
                    case 5: return [2];
                }
            });
        });
    };
    AlarmRunnerService.prototype.createTaskManager = function () {
        return new manager_1.TaskManager(planned_task_1.PlanningType.Scheduled, this.taskStore, this.foregroundChecker, this.timeOffset, this.currentTime);
    };
    AlarmRunnerService.prototype.initializeExecutionWindow = function (timeout) {
        var _this = this;
        this.wakeLock.acquire(timeout);
        var executionTimeout = timeout - TIMEOUT_EVENT_OFFSET;
        var startEvent = events_1.createEvent(events_1.TaskDispatcherEvent.TaskExecutionStarted, {
            expirationTimestamp: this.getExpirationTimestamp(executionTimeout),
        });
        var id = startEvent.id;
        var timeoutEvent = events_1.createEvent(events_1.TaskDispatcherEvent.TaskExecutionTimedOut, {
            id: id,
        });
        this.logger.info("Execution will timeout in " + executionTimeout + ", for tasks running with execution id: " + id);
        this.timeoutId = setTimeout(function () {
            events_1.emit(timeoutEvent);
            _this.logger.warn("Execution timed out for tasks running with execution id: " + id);
        }, executionTimeout);
        return startEvent;
    };
    AlarmRunnerService.prototype.calculateTimeout = function (taskPlanner) {
        return __awaiter(this, void 0, void 0, function () {
            var nextExecutionTime;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, taskPlanner.nextInterval()];
                    case 1:
                        nextExecutionTime = _a.sent();
                        return [2, Math.max(nextExecutionTime, MIN_TIMEOUT)];
                }
            });
        });
    };
    AlarmRunnerService.prototype.getExpirationTimestamp = function (timeout) {
        return time_1.now() + timeout;
    };
    AlarmRunnerService.prototype.gracefullyStop = function () {
        this.moveToBackground();
        if (this.started) {
            this.logger.debug("Stopping service");
            this.killWithFire();
            this.started = false;
        }
        if (this.timeoutId) {
            clearTimeout(this.timeoutId);
        }
        if (this.wakeLock.isHeld()) {
            this.wakeLock.release();
            this.logger.debug("Lock released");
        }
    };
    AlarmRunnerService.prototype.moveToBackground = function () {
        if (!this.inForeground) {
            return;
        }
        var andRemoveNotification = true;
        this.nativeService.stopForeground(andRemoveNotification);
        this.inForeground = false;
        this.logger.debug("Running in background");
    };
    AlarmRunnerService.prototype.killWithFire = function () {
        var startId = 1;
        var last = false;
        while (!last) {
            last = this.nativeService.stopSelfResult(startId);
            if (!last) {
                startId++;
            }
        }
        this.logger.info("Done running (startId=" + startId + ")");
    };
    return AlarmRunnerService;
}());
exports.AlarmRunnerService = AlarmRunnerService;
function alarmRunnerWakeLock(context) {
    var wakeLockName = "TaskDispatcher::AlarmRunnerWakeLock";
    var powerManager = context.getSystemService(android.content.Context.POWER_SERVICE);
    return powerManager.newWakeLock(android.os.PowerManager.PARTIAL_WAKE_LOCK, wakeLockName);
}
var _alarmRunnerService;
function getAlarmRunnerService() {
    if (!_alarmRunnerService) {
        notification_manager_android_1.setupNotificationChannels(application_1.android.context);
        _alarmRunnerService = new AlarmRunnerService();
    }
    return _alarmRunnerService;
}
exports.getAlarmRunnerService = getAlarmRunnerService;
//# sourceMappingURL=runner-service.android.js.map