"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var logger_1 = require("../../../../../../utils/logger");
var await_lock_1 = require("await-lock");
var manager_android_1 = require("./manager.android");
var manager_android_2 = require("../watchdog/manager.android");
var planned_tasks_store_1 = require("../../../../../../persistence/planned-tasks-store");
var planned_task_1 = require("../../../../../planner/planned-task");
var time_1 = require("../../../../../../utils/time");
var MIN_ALARM_INTERVAL = 60000;
var AndroidAlarmScheduler = (function () {
    function AndroidAlarmScheduler(alarmManager, watchdogManager, plannedTaskStore) {
        if (alarmManager === void 0) { alarmManager = new manager_android_1.AndroidAlarmManager(); }
        if (watchdogManager === void 0) { watchdogManager = new manager_android_2.WatchdogManager(); }
        if (plannedTaskStore === void 0) { plannedTaskStore = planned_tasks_store_1.plannedTasksDB; }
        this.alarmManager = alarmManager;
        this.watchdogManager = watchdogManager;
        this.plannedTaskStore = plannedTaskStore;
        this.tag = "AndroidAlarmScheduler";
        this.logger = logger_1.getLogger(this.tag);
        this.lock = new await_lock_1.default();
    }
    AndroidAlarmScheduler.prototype.setup = function () {
        return __awaiter(this, void 0, void 0, function () {
            var plannedTasks;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (this.alarmManager.alarmUp && this.watchdogManager.alarmUp) {
                            return [2];
                        }
                        return [4, this.plannedTaskStore.getAllSortedByNextRun(planned_task_1.PlanningType.Scheduled)];
                    case 1:
                        plannedTasks = _a.sent();
                        if (plannedTasks.length > 0) {
                            if (!this.alarmManager.alarmUp) {
                                this.logger.info("Alarm was not up! Scheduling...");
                                this.alarmManager.set(calculateAlarmInterval(plannedTasks[0]));
                            }
                            if (!this.watchdogManager.alarmUp) {
                                this.logger.info("Watchdog was not up! Initializing...");
                                this.watchdogManager.set();
                            }
                        }
                        return [2];
                }
            });
        });
    };
    AndroidAlarmScheduler.prototype.schedule = function (runnableTask) {
        return __awaiter(this, void 0, void 0, function () {
            var plannedTask;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.lock.acquireAsync()];
                    case 1:
                        _a.sent();
                        return [4, this.onSchedule(runnableTask)];
                    case 2:
                        plannedTask = _a.sent();
                        this.lock.release();
                        return [2, plannedTask];
                }
            });
        });
    };
    AndroidAlarmScheduler.prototype.cancel = function (id) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.lock.acquireAsync()];
                    case 1:
                        _a.sent();
                        return [4, this.onCancel(id)];
                    case 2:
                        _a.sent();
                        this.lock.release();
                        return [2];
                }
            });
        });
    };
    AndroidAlarmScheduler.prototype.onSchedule = function (runnableTask) {
        return __awaiter(this, void 0, void 0, function () {
            var possibleExisting, allTasks, currentMillis, plannedTask;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.plannedTaskStore.get(runnableTask)];
                    case 1:
                        possibleExisting = _a.sent();
                        if (possibleExisting) {
                            return [2, possibleExisting];
                        }
                        return [4, this.plannedTaskStore.getAllSortedByNextRun(planned_task_1.PlanningType.Scheduled)];
                    case 2:
                        allTasks = _a.sent();
                        currentMillis = time_1.now();
                        plannedTask = new planned_task_1.PlannedTask(planned_task_1.PlanningType.Scheduled, planned_task_1.SchedulerType.Alarm, runnableTask);
                        if (allTasks.length === 0 ||
                            allTasks[0].nextRun(currentMillis) > plannedTask.nextRun(currentMillis)) {
                            this.alarmManager.set(calculateAlarmInterval(plannedTask));
                            if (!this.watchdogManager.alarmUp) {
                                this.watchdogManager.set();
                            }
                        }
                        return [4, this.plannedTaskStore.insert(plannedTask)];
                    case 3:
                        _a.sent();
                        this.logger.info("Task scheduled: " + JSON.stringify(plannedTask));
                        return [2, plannedTask];
                }
            });
        });
    };
    AndroidAlarmScheduler.prototype.onCancel = function (id) {
        return __awaiter(this, void 0, void 0, function () {
            var possibleExisting, allTasks;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.plannedTaskStore.get(id)];
                    case 1:
                        possibleExisting = _a.sent();
                        if (!possibleExisting) {
                            return [2];
                        }
                        return [4, this.plannedTaskStore.getAllSortedByNextRun(planned_task_1.PlanningType.Scheduled)];
                    case 2:
                        allTasks = _a.sent();
                        if (allTasks.length === 1) {
                            this.alarmManager.cancel();
                            this.watchdogManager.cancel();
                        }
                        else if (allTasks[0].id === possibleExisting.id &&
                            !areExecutedByTheSameAlarm(allTasks[1], possibleExisting)) {
                            this.alarmManager.set(calculateAlarmInterval(allTasks[1]));
                        }
                        return [4, this.plannedTaskStore.delete(id)];
                    case 3:
                        _a.sent();
                        this.logger.info("Task with id=" + id + " has been canceled");
                        return [2];
                }
            });
        });
    };
    return AndroidAlarmScheduler;
}());
exports.AndroidAlarmScheduler = AndroidAlarmScheduler;
function calculateAlarmInterval(plannedTask) {
    var nextRun = plannedTask.nextRun();
    return nextRun > MIN_ALARM_INTERVAL ? nextRun : MIN_ALARM_INTERVAL;
}
function areExecutedByTheSameAlarm(pt1, pt2) {
    var currentTime = time_1.now();
    var nextRunDiff = Math.abs(pt1.nextRun(currentTime) - pt2.nextRun(currentTime));
    return nextRunDiff < MIN_ALARM_INTERVAL / 2;
}
//# sourceMappingURL=scheduler.android.js.map