"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var events_1 = require("../events");
var logger_1 = require("../utils/logger");
var task_chain_1 = require("./task-chain");
var time_1 = require("../utils/time");
var serialization_1 = require("../utils/serialization");
var Task = (function () {
    function Task(name, taskConfig) {
        if (taskConfig === void 0) { taskConfig = {}; }
        this._name = name;
        this._executionHistory = new Set();
        this._cancelFunctions = new Map();
        this.configureTask(taskConfig);
    }
    Object.defineProperty(Task.prototype, "name", {
        get: function () {
            return this._name;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Task.prototype, "outputEventNames", {
        get: function () {
            return this._taskConfig.outputEventNames;
        },
        enumerable: true,
        configurable: true
    });
    Task.prototype.run = function (taskParams, invocationEvent) {
        return __awaiter(this, void 0, void 0, function () {
            var outcome, err_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (this._invocationEvent && !this.isDone()) {
                            this.cancelParallelInvocation(taskParams, invocationEvent);
                            return [2];
                        }
                        this._taskParams = taskParams;
                        this._invocationEvent = invocationEvent;
                        this.log("Run triggered by " + invocationEvent.name + " event " + (invocationEvent.expirationTimestamp !== -1
                            ? "with " + invocationEvent.expirationTimestamp + " expirationTimestamp"
                            : ""));
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 4, , 5]);
                        return [4, this.checkIfCanRun()];
                    case 2:
                        _a.sent();
                        return [4, this.onRun(taskParams, invocationEvent)];
                    case 3:
                        outcome = _a.sent();
                        if (!this.isDone()) {
                            this.processTaskOutcome(outcome);
                        }
                        return [3, 5];
                    case 4:
                        err_1 = _a.sent();
                        this.getLogger().error("Execution failed with params " + JSON.stringify(taskParams) + " and invocation event " + JSON.stringify(invocationEvent) + ": " + (err_1.stack ? err_1.stack : err_1));
                        this.emitEndEvent(task_chain_1.TaskResultStatus.Error, err_1);
                        throw err_1;
                    case 5:
                        this.removeCancelFunction();
                        return [2];
                }
            });
        });
    };
    Task.prototype.runsInBackground = function () {
        return !this._taskConfig.foreground;
    };
    Task.prototype.cancel = function () {
        this.emitEndEvent(task_chain_1.TaskResultStatus.Cancelled);
        if (this._cancelFunctions.has(this._invocationEvent.id)) {
            var cancelFunction = this._cancelFunctions.get(this._invocationEvent.id);
            cancelFunction();
            this.removeCancelFunction();
        }
        this.log("Cancelled");
    };
    Task.prototype.checkIfCanRun = function () {
        return Promise.resolve();
    };
    Task.prototype.prepare = function () {
        return Promise.resolve();
    };
    Task.prototype.runAgainIn = function (seconds, params) {
        defer(this.name, seconds, params ? params : this._taskParams);
        this.log("Will run again in " + seconds + " s");
    };
    Task.prototype.setCancelFunction = function (f) {
        if (this.isDone()) {
            f();
            return;
        }
        this._cancelFunctions.set(this._invocationEvent.id, f);
    };
    Task.prototype.log = function (message) {
        this.getLogger().info(message + " (invocationId=" + this._invocationEvent.id + ")");
    };
    Task.prototype.remainingTime = function () {
        if (this._invocationEvent.expirationTimestamp === -1) {
            return -1;
        }
        var timeForExpiration = this._invocationEvent.expirationTimestamp - time_1.now();
        if (this.outputEventNames.some(events_1.hasListeners)) {
            timeForExpiration *= 0.9;
        }
        return Math.floor(timeForExpiration);
    };
    Task.prototype.configureTask = function (taskConfig) {
        this._taskConfig = {
            foreground: taskConfig.foreground,
            outputEventNames: taskConfig.outputEventNames && taskConfig.outputEventNames.length > 0
                ? taskConfig.outputEventNames
                : [this.name + "Finished"],
        };
    };
    Task.prototype.cancelParallelInvocation = function (taskParams, invocationEvent) {
        this.getLogger().warn("Multiple executions of a task cannot run concurrently. Warning triggered with: params=" + JSON.stringify(taskParams) + " and invocationEvent=" + JSON.stringify(invocationEvent));
        var reason = new Error("Concurrent Execution: Executing multiple instances of a task concurrently is not allowed");
        this.emitEndEvent(task_chain_1.TaskResultStatus.Cancelled, reason, invocationEvent.id);
    };
    Task.prototype.processTaskOutcome = function (outcome) {
        if (this.outputEventNames.length > 1 && (!outcome || !outcome.eventName)) {
            this.getLogger().error("Task was declared to have more than one possible output event, but outcome did not specify any.\n        Please, check if you forgot to fill eventName field in any of the outputs of your conditional branches.\n        Received outcome: " + JSON.stringify(outcome));
            var reason = new Error("Cannot choose which event to emit");
            this.emitEndEvent(task_chain_1.TaskResultStatus.Error, reason);
            return;
        }
        var outputEvent = !outcome || !outcome.eventName
            ? this.outputEventNames[0]
            : outcome.eventName;
        if (!outcome || !outcome.result) {
            this.done(outputEvent);
        }
        else {
            this.done(outputEvent, outcome.result);
        }
    };
    Task.prototype.done = function (eventName, result) {
        if (this.isDone()) {
            return;
        }
        this.markAsDone();
        if (!this.outputEventNames.find(function (name) { return name === eventName; })) {
            this.getLogger().warn("About to emit an event (" + eventName + ") not declared in the list of possible output events of this task.\n      This can lead to unintended behaviors. Perhaps you have forgot to include it in the task configuration or there is a typo");
        }
        if (!events_1.hasListeners(eventName)) {
            this.emitEndEvent(task_chain_1.TaskResultStatus.Ok);
            return;
        }
        var data = result
            ? typeof result === "object"
                ? serialization_1.flatten(result)
                : { result: result }
            : {};
        events_1.emit({
            name: eventName,
            id: this._invocationEvent.id,
            expirationTimestamp: this._invocationEvent.expirationTimestamp,
            data: data,
        });
        this.log("Finished running with " + eventName + " event");
    };
    Task.prototype.isDone = function () {
        return this._executionHistory.has(this._invocationEvent.id);
    };
    Task.prototype.markAsDone = function (invocationId) {
        var id = invocationId ? invocationId : this._invocationEvent.id;
        this._executionHistory.add(id);
    };
    Task.prototype.emitEndEvent = function (status, err, invocationId) {
        var id = invocationId ? invocationId : this._invocationEvent.id;
        this.markAsDone(id);
        task_chain_1.TaskChain.finalize(id, status, err);
    };
    Task.prototype.removeCancelFunction = function () {
        this._cancelFunctions.delete(this._invocationEvent.id);
    };
    Task.prototype.getLogger = function () {
        if (!this._logger) {
            this._logger = logger_1.getLogger("Task (" + this._name + ")");
        }
        return this._logger;
    };
    return Task;
}());
exports.Task = Task;
var defer;
function setTaskDeferrer(taskDeferrer) {
    defer = taskDeferrer;
}
exports.setTaskDeferrer = setTaskDeferrer;
//# sourceMappingURL=task.js.map